#include "Operator.h"
#include "ElInfo.h"
#include "Assembler.h"
#include "FixVec.h"
#include "DOFVector.h"
#include "Quadrature.h"
#include "OpenMP.h"

namespace AMDiS {

  const Flag Operator::MATRIX_OPERATOR = 1;
  const Flag Operator::VECTOR_OPERATOR = 2;


  int Operator::getQuadratureDegree(int order, FirstOrderType firstOrderType) 
  {
    std::vector<OperatorTerm*>* terms = NULL;
    int myRank = omp_get_thread_num();

    switch(order) {
    case 0:
      terms = &zeroOrder[myRank];
      break;
    case 1:
      if (firstOrderType == GRD_PHI)
	terms = &firstOrderGrdPhi[myRank];
      else 
	terms = &firstOrderGrdPsi[myRank];
      break;
    case 2:
      terms = &secondOrder[myRank];
      break;
    }

    const BasisFunction *psi = rowFESpace->getBasisFcts();
    const BasisFunction *phi = colFESpace->getBasisFcts();

    int psiDegree = psi->getDegree();
    int phiDegree = phi->getDegree();
    int maxTermDegree = 0;

    for (int i = 0; i < static_cast<int>(terms->size()); i++) {
      OperatorTerm *term = (*terms)[i];
      maxTermDegree = max(maxTermDegree, term->degree);
    }
   
    return psiDegree + phiDegree - order + maxTermDegree;
  }


  Operator::Operator(Flag operatorType,
		     const FiniteElemSpace *row,
		     const FiniteElemSpace *col)
    : rowFESpace(row), 
      colFESpace(col ? col : row),
      type(operatorType), 
      fillFlag(Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS | 
	       Mesh::FILL_DET | Mesh::FILL_GRD_LAMBDA),
      needDualTraverse(false),
      uhOld(NULL),
      optimized(true)
  {
    int maxThreads = omp_get_overall_max_threads();

    assembler.resize(maxThreads);
    secondOrder.resize(maxThreads);
    firstOrderGrdPsi.resize(maxThreads);
    firstOrderGrdPhi.resize(maxThreads);
    zeroOrder.resize(maxThreads);

    for (int i = 0; i < maxThreads; i++) {
      assembler[i] = NULL;
      secondOrder[i].resize(0);
      firstOrderGrdPsi[i].resize(0);
      firstOrderGrdPhi[i].resize(0);
      zeroOrder[i].resize(0);
    }

    nRow = rowFESpace->getBasisFcts()->getNumber();
    nCol = colFESpace->getBasisFcts()->getNumber();
  }


  void Operator::setUhOld(const DOFVectorBase<double> *vec)
  {
    uhOld = vec;
    auxFeSpaces.insert(vec->getFESpace());
  }


  void Operator::getElementMatrix(const ElInfo *elInfo, 
				  ElementMatrix& userMat, 
				  double factor)
  {
    int myRank = omp_get_thread_num();

    if (!assembler[myRank])
      initAssembler(myRank, NULL, NULL, NULL, NULL);

    assembler[myRank]->calculateElementMatrix(elInfo, userMat, factor);
  }


  void Operator::getElementMatrix(const ElInfo *rowElInfo, const ElInfo *colElInfo,
				  const ElInfo *smallElInfo, const ElInfo *largeElInfo,
				  ElementMatrix& userMat, 
				  double factor)
  {
    int myRank = omp_get_thread_num();

    if (!assembler[myRank])
      initAssembler(myRank, NULL, NULL, NULL, NULL);

    assembler[myRank]->calculateElementMatrix(rowElInfo, colElInfo, 
					      smallElInfo, largeElInfo,
    					      userMat, factor);
  }


  void Operator::getElementVector(const ElInfo *elInfo, 
				  ElementVector& userVec, 
				  double factor)
  {
    int myRank = omp_get_thread_num();

    if (!assembler[myRank])
      initAssembler(myRank, NULL, NULL, NULL, NULL);

    assembler[myRank]->calculateElementVector(elInfo, userVec, factor);
  }

  void Operator::getElementVector(const ElInfo *mainElInfo, const ElInfo *auxElInfo,
				  const ElInfo *smallElInfo, const ElInfo *largeElInfo,
				  ElementVector& userVec,
				  double factor)
  {
    int myRank = omp_get_thread_num();

    if (!assembler[myRank])
      initAssembler(myRank, NULL, NULL, NULL, NULL);

    assembler[myRank]->calculateElementVector(mainElInfo, auxElInfo, 
					      smallElInfo, largeElInfo,
					      userVec, factor);
  }

  void Operator::initAssembler(int rank,
			       Quadrature *quad2,
			       Quadrature *quad1GrdPsi,
			       Quadrature *quad1GrdPhi,
			       Quadrature *quad0) 
  {    
#ifdef _OPENMP
#pragma omp critical (initAssembler)
#endif
      {
	if (optimized)
	  assembler[rank] = 
	    new OptimizedAssembler(this, quad2, quad1GrdPsi, quad1GrdPhi, quad0,
				   rowFESpace, colFESpace);
	else
	  assembler[rank] = 
	    new StandardAssembler(this, quad2, quad1GrdPsi, quad1GrdPhi, quad0,
				  rowFESpace, colFESpace);	
      }
  }

  void Operator::finishAssembling()
  {
    assembler[omp_get_thread_num()]->finishAssembling();
  }


  Assembler *Operator::getAssembler(int rank) 
  {
    if (!assembler[rank])
      initAssembler(rank, NULL, NULL, NULL, NULL);    

    return assembler[rank];
  }

  void Operator::setAssembler(int rank, Assembler *a)
  {
    assembler[rank] = a; 
  }

  void Operator::weakEvalSecondOrder(const std::vector<WorldVector<double> > &grdUhAtQP,
				     std::vector<WorldVector<double> > &result) const
  {
    int myRank = omp_get_thread_num();
    
    for (std::vector<OperatorTerm*>::const_iterator termIt = secondOrder[myRank].begin(); 
	 termIt != secondOrder[myRank].end(); 
	 ++termIt)
      static_cast<SecondOrderTerm*>(*termIt)->weakEval(grdUhAtQP, result);
  }
 
}
