#include "AMDiS.h"
#include "ParallelProblem.h"
#include "mpi.h"

using namespace std;
using namespace AMDiS;

// ===========================================================================
// ===== function definitions ================================================
// ===========================================================================

/// Dirichlet boundary function
class G : public AbstractFunction<double, WorldVector<double> >,
	  public TimedObject
{
public:
  /// Implementation of AbstractFunction::operator().
  double operator()(const WorldVector<double>& argX) const 
  {
    WorldVector<double> x = argX;
    int dim = x.getSize();
    for (int i = 0; i < dim; i++)
      x[i] -= *timePtr;
    return sin(M_PI*(*timePtr)) * exp(-10.0*(x*x));
  }
}

/// RHS function
class F : public AbstractFunction<double, WorldVector<double> >,
	  public TimedObject
{
public:
  F(int degree) : AbstractFunction<double, WorldVector<double> >(degree) {}

  /// Implementation of AbstractFunction::operator().
  double operator()(const WorldVector<double>& argX) const 
  {
    WorldVector<double> x = argX;
    int dim = x.getSize();
    for (int i = 0; i < dim; i++)
      x[i] -= *timePtr;
    double r2 = (x*x);
    double ux = sin(M_PI * (*timePtr)) * exp(-10.0*r2);
    double ut = M_PI * cos(M_PI*(*timePtr)) * exp(-10.0*r2);
    return ut -(400.0*r2 - 20.0*dim)*ux;
  }
};

// ===========================================================================
// ===== instationary problem ================================================
// ===========================================================================

/// Instationary problem
class Heat : public ProblemInstatScal
{
public:
  /// Constructor
  Heat(ProblemScal *heatSpace)
    : ProblemInstatScal("heat", heatSpace)
  {}

  // ===== ProblemInstatBase methods ===================================

  /// set the time in all needed functions!
  void setTime(AdaptInfo *adaptInfo) 
  {
    time = adaptInfo->getTime();
    tau1 = 1.0 / adaptInfo->getTimestep();    
  }

  // ===== initial problem methods =====================================

  /// Used by \ref problemInitial to solve the system of the initial problem
  void solve(AdaptInfo *adaptInfo) 
  {
    problemStat->getMesh()->dofCompress();
    time = adaptInfo->getStartTime();
    problemStat->getSolution()->interpol(boundaryFct);
  }

  /// Used by \ref problemInitial to do error estimation for the initial problem.
  void estimate(AdaptInfo *adaptInfo) 
  {
    double errMax, errSum;
    time = adaptInfo->getStartTime();
    errSum = Error<double>::L2Err(*boundaryFct,
				  *(problemStat->getSolution()), 0, &errMax, false);
    adaptInfo->setEstSum(errSum, 0);
  }

  // ===== setting methods ===============================================

  /// Sets \ref boundaryFct;
  void setBoundaryFct(AbstractFunction<double, WorldVector<double> > *fct) 
  {
    boundaryFct = fct;
  } 

  // ===== getting methods ===============================================

  /// Returns pointer to \ref tau1
  double *getTau1Ptr() 
  { 
    return &tau1; 
  }

  /// Returns pointer to \ref time.
  double *getTimePtr() 
  { 
    return &time; 
  }

  /// Returns \ref boundaryFct;
  AbstractFunction<double, WorldVector<double> > *getBoundaryFct() 
  {
    return boundaryFct;
  }


//   void closeTimestep(AdaptInfo *adaptInfo) {
//     ParallelProblem::writeRankMacroAndValues(problemStat->getSolution(),
// 					     "output/debug_rank", 
// 					     adaptInfo->getTime());
//   };

private:
  /// 1.0 / timestep
  double tau1;

  /// time 
  double time;

  /// Pointer to boundary function. Needed for initial problem.
  AbstractFunction<double, WorldVector<double> > *boundaryFct;
};

// ===========================================================================
// ===== main program ========================================================
// ===========================================================================

int main(int argc, char** argv)
{
  MPI::Init(argc, argv);

  // ===== check for init file =====
  TEST_EXIT(argc >= 2)("usage: heat initfile\n");

  // ===== init parameters =====
  Parameters::init(false, argv[1]);

  // ===== create and init stationary problem =====
  ProblemScal *heatSpace = new ProblemScal("heat->space");
  heatSpace->initialize(INIT_ALL);

  // ===== create instationary problem =====
  Heat *heat = new Heat(heatSpace);
  heat->initialize(INIT_ALL);

  // create adapt info
  AdaptInfo *adaptInfo = new AdaptInfo("heat->adapt");

  // create initial adapt info
  AdaptInfo *adaptInfoInitial = new AdaptInfo("heat->initial->adapt");

  // create parallel problem
  std::vector<DOFVector<double>*> vectors;
  ParallelProblemScal parallelheat("heat->parallel", heatSpace, heat, vectors);

  // create instationary adapt
  AdaptInstationary *adaptInstat =
    new AdaptInstationary("heat->adapt",
			  &parallelheat,
			  adaptInfo,
			  &parallelheat,
			  adaptInfoInitial);

  // ===== create rhs functions =====
  int degree = heatSpace->getFESpace()->getBasisFcts()->getDegree();
  F *rhsFct = new F(degree);
  rhsFct->setTimePtr(heat->getTimePtr());

  // ===== create operators =====
  double one = 1.0;
  double zero = 0.0;

  // create laplace
  Operator *A = new Operator(Operator::MATRIX_OPERATOR | 
			     Operator::VECTOR_OPERATOR,
			     heatSpace->getFESpace());

  A->addSecondOrderTerm(new Laplace_SOT);

  A->setUhOld(heat->getOldSolution());

  heatSpace->addMatrixOperator(A, &one, &one);

  // create zero order operator
  Operator *C = new Operator(Operator::MATRIX_OPERATOR | 
			     Operator::VECTOR_OPERATOR,
			     heatSpace->getFESpace());

  C->addZeroOrderTerm(new Simple_ZOT);

  C->setUhOld(heat->getOldSolution());

  heatSpace->addMatrixOperator(C, heat->getTau1Ptr(), heat->getTau1Ptr());
  heatSpace->addVectorOperator(C, heat->getTau1Ptr(), heat->getTau1Ptr());

  // create RHS operator
  Operator *F = new Operator(Operator::VECTOR_OPERATOR,
			     heatSpace->getFESpace());

  F->addZeroOrderTerm(new CoordsAtQP_ZOT(rhsFct));

  heatSpace->addVectorOperator(F);

  // ===== create boundary functions =====
  G *boundaryFct = new G;
  boundaryFct->setTimePtr(heat->getTimePtr());
  heat->setBoundaryFct(boundaryFct);

  heatSpace->addDirichletBC(DIRICHLET, boundaryFct);



//   // ===== start adaption loop =====
  parallelheat.initParallelization(adaptInfo);

  adaptInstat->adapt();

  parallelheat.exitParallelization(adaptInfo);

  MPI::Finalize();
}
